/*___________________________________________________________________________

[ pam_capability ] the capabilities pam module

Copyright (C) 2002 Erik M. A. Kline

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Please see the file "GPL" in the "meta" subdirectory of the top level source
directory, if you have received this program with source code.

For further information or concerns please contact:

    Erik Kline
    ekline@ekline.com, or
    erik@alum.mit.edu
  ___________________________________________________________________________
*/

#define PAM_SM_SESSION

#include <security/_pam_aconf.h>
#include <security/_pam_macros.h>
#include <security/pam_modules.h>
#include <sys/capability.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <syslog.h>
#include <string.h>
#include <error.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

///////////////////////////////////////////////////////////////////////////
// local defines, typedefs, and prototypes ////////////////////////////////
///////////////////////////////////////////////////////////////////////////

#define	MAX_ROLENAME	35
#define	MAX_CAPTEXT	2048
#define	MAX_USERNAME	35
#define	MAX_GROUPNAME	35
#define	MAX_LINE	2048

#define	CAP_CONF	CONF_FILE

#define	CONF_BAD_PERMS	( S_IWGRP | S_IXGRP | S_IWOTH | S_IXOTH )
#define	WHITESPACE	" \t\r\n"
// #define	CAP_EXTRA	" cap_setpcap-eip cap_setuid+ep cap_setgid+ep"
#define	CAP_EXTRA	"cap_setpcap-eip"

#define	APPEND( x, y, m )	do {                                          \
					strncpy(x+strlen(x), y, m-strlen(x)); \
				} while(0);


// role struct and routines
typedef struct pamcap_role {
	char rolename[MAX_ROLENAME];
	char capabilities[MAX_CAPTEXT];
	struct pamcap_role * next;
} pamcap_role_t;

void pamcap_role_init( pamcap_role_t * role );
void pamcap_role_freelist( pamcap_role_t * roletop );
pamcap_role_t * pamcap_role_findlist( pamcap_role_t * roletop,
					   const char * rolename );
int pamcap_role_appendlist( pamcap_role_t ** roletop, pamcap_role_t * new );
int pamcap_role_verify( pamcap_role_t * role );


// user struct and routines
typedef struct pamcap_user {
	char username[MAX_USERNAME];
	char rolename[MAX_ROLENAME];
	struct pamcap_user * next;
} pamcap_user_t;

void pamcap_user_init( pamcap_user_t * user );
void pamcap_user_freelist( pamcap_user_t * usertop );
pamcap_user_t * pamcap_user_findlist( pamcap_user_t * usertop,
				      const char * username );
int pamcap_user_appendlist( pamcap_user_t ** usertop, pamcap_user_t * new );
int pamcap_user_verify( pamcap_user_t * user );


// user struct and routines
typedef struct pamcap_group {
	char groupname[MAX_USERNAME];
	char rolename[MAX_ROLENAME];
	struct pamcap_group * next;
} pamcap_group_t;

void pamcap_group_init( pamcap_group_t * group );
void pamcap_group_freelist( pamcap_group_t * grouptop );
pamcap_group_t * pamcap_group_findlist( pamcap_group_t * grouptop,
				        const char * groupname );
int pamcap_group_appendlist( pamcap_group_t ** grouptop, pamcap_group_t * new );
pamcap_group_t * pamcap_group_findbyuser( pamcap_group_t * grouptop,
					  const char * username );
int pamcap_group_verify( pamcap_group_t * group );


// obj struct and routines
typedef struct pamcap_obj {
	int debug;
	const char * conf;
	const char * append;
	pamcap_role_t  * tbl_roles;
	pamcap_user_t  * tbl_users;
	pamcap_group_t * tbl_groups;
} pamcap_obj_t;

int pamcap_obj_set_default( pamcap_obj_t * obj );
int pamcap_obj_parse( pamcap_obj_t * obj, int argc, const char ** argv );


int pamcap_set_by_user( const char * username, pamcap_obj_t * obj );
void pamcap_set_default( pamcap_obj_t * obj );

int pamcap_conf_check( pamcap_obj_t * obj );
int pamcap_conf_parse( pamcap_obj_t * obj );

///////////////////////////////////////////////////////////////////////////
// PAM external calls /////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

PAM_EXTERN int pam_sm_open_session( pam_handle_t *pamh, int flags,
				    int argc, const char **argv )
{
	int rval = PAM_SESSION_ERR;
	cap_t caps = NULL;
	char * captext = NULL;
	const char * username = NULL;
	pamcap_obj_t obj;

	openlog( "pam_capability", LOG_PID, LOG_AUTHPRIV );
	pamcap_obj_set_default( &obj );
	
	if ( pamcap_obj_parse(&obj, argc, argv) != 0 ) {
		syslog( LOG_ERR, "couldn't parse module options" );
		goto __out;
	}

	if ( (caps = cap_get_proc()) == NULL ) {
		syslog( LOG_ERR, "couldn't get proc capabilities" );
		goto __out;
	}

	// there is regrettably no way to find out what your own
	// capabilities are in userland, so we perform a hack:
	if ( cap_set_proc(caps) != 0 ) {
		syslog( LOG_ERR, "don't seem to have CAP_SETPCAP set" );
		// bail if we aren't using an appropriately modified kernel
		rval = PAM_SUCCESS;
		goto __short_cut;
	}

	if ( (rval = pam_get_user(pamh, &username, NULL)) != PAM_SUCCESS ) {
		syslog( LOG_ERR, "can't get username: %s",
		        pam_strerror(pamh, rval) );
		goto __out;
	}

	if ( username == NULL || *username == '\0' ) {
		syslog( LOG_ERR, "invalid (null) username" );
		goto __out;
	}

	// if we can't find anything useful in the user and role
	// configuration files we should apply the default policy
	if ( pamcap_set_by_user(username, &obj) != 0 ) {
		if ( obj.debug )
			syslog( LOG_INFO, "using default policy for user %s",
				username );
		pamcap_set_default( &obj );
	}

__short_cut:
	if ( obj.debug )
		syslog( LOG_INFO, "session open" );

	rval = PAM_SUCCESS;
__out:
	if ( rval != PAM_SUCCESS )
		pamcap_set_default( &obj );

	if ( caps != NULL )
		cap_free( (void *)caps );

	if ( obj.debug ) {
		caps = cap_get_proc();
		captext = cap_to_text( caps, NULL );
		syslog( LOG_INFO, "session capabilities: %s", captext );
		cap_free( (void *)captext );
		cap_free( (void *)caps );
	}

	closelog();
	return rval;
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh, int flags,
				    int argc, const char **argv)
{
	pamcap_obj_t obj;

	openlog( "pam_capability", LOG_PID, LOG_AUTHPRIV );
	pamcap_obj_set_default( &obj );
	
	if ( pamcap_obj_parse(&obj, argc, argv) != 0 )
		syslog( LOG_ERR, "couldn't parse module options" );

	if ( obj.debug )
		syslog( LOG_INFO, "session closed" );

	closelog();
	return PAM_SUCCESS;
}

#ifdef PAM_STATIC

/*
 * static module data
 */

struct pam_module _pam_capability_modstruct = {
	"pam_capability",
	NULL,
	NULL,
	pam_sm_acct_mgmt,
	NULL,
	NULL,
	NULL
};

#endif

///////////////////////////////////////////////////////////////////////////
// local routines /////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

// "pamcap_option_t" routines /////////////////////////////////////////////
int pamcap_obj_set_default( pamcap_obj_t * obj )
{
	if ( obj == NULL )
		return -1;

	obj->debug      = 0;
	obj->conf       = CAP_CONF;
	obj->append     = NULL;
	obj->tbl_roles  = NULL;
	obj->tbl_users  = NULL;
	obj->tbl_groups = NULL;

	return 0;
}

int pamcap_obj_parse( pamcap_obj_t * obj,
				  int argc, const char ** argv )
{
	int i;

	if ( obj == NULL )
		return -1;

	for ( i = 0; i < argc; i++ ) {
		if ( strcmp(argv[i], "debug") == 0 )
			obj->debug = 1;
		else if ( strncmp(argv[i], "conf=", 5) == 0 )
			obj->conf = argv[i] + 5;
		else if ( strncmp(argv[i], "append=", 7) == 0 )
			obj->append = argv[i] + 7;
		else
			syslog( LOG_INFO, "ignoring unknown option <%s>",
				argv[i] );
	}

	if ( obj->debug )
		syslog( LOG_NOTICE, "debug=%s, conf=%s append=%s",
			obj->debug ? "TRUE" : "FALSE", obj->conf,
			obj->append );

	return 0;
}

// "pamcap_role_t" routines ///////////////////////////////////////////////
void pamcap_role_init( pamcap_role_t * role )
{
	if ( role == NULL )
		return;

	memset( role->rolename, 0, MAX_ROLENAME );
	memset( role->capabilities, 0, MAX_CAPTEXT );
	role->next = NULL;
}

void pamcap_role_freelist( pamcap_role_t * roletop )
{
	pamcap_role_t * ptr1, * ptr2;

	for ( ptr1 = roletop; ptr1 != NULL; ptr1 = ptr2 ) {
		ptr2 = ptr1->next;
		free( ptr1 );
	}
}

pamcap_role_t * pamcap_role_findlist( pamcap_role_t * roletop,
					   const char * rolename )
{
	pamcap_role_t * tmp = NULL;

	for ( tmp = roletop; tmp != NULL; tmp = tmp->next )
		if ( strcmp(tmp->rolename, rolename) == 0 )
			break;

	return tmp;
}

int pamcap_role_appendlist( pamcap_role_t ** roletop, pamcap_role_t * new )
{
	pamcap_role_t * rp;
	cap_t caps = NULL;

	if ( roletop == NULL )
		return -1;

	if ( new == NULL )
		return 0;

	// enforce uniqueness
	if ( pamcap_role_findlist(*roletop, new->rolename) != NULL ) {
		syslog( LOG_ERR, "duplicate role (%s) ignored", new->rolename );
		return -1;
	}

	// verify capabilities compiles
	if ( (caps = cap_from_text(new->capabilities)) == NULL ) {
		syslog( LOG_ERR, "invalid capability listing (%s) ignored",
			new->capabilities );
		return -1;
	} 

	cap_free( caps );

	if ( *roletop == NULL )
		*roletop = new;
	else {
		for ( rp = *roletop; rp->next != NULL; rp = rp->next )
			;;
		rp->next = new;
	}

	return 0;
}

int pamcap_role_verify( pamcap_role_t * role )
{
	cap_t caps = NULL;

	if ( role == NULL )
		return -1;

	if ( strlen(role->rolename) == 0 ||
	     strlen(role->rolename) > MAX_ROLENAME )
		return -1;

	if ( (caps = cap_from_text(role->capabilities)) == NULL )
		return -1;

	cap_free( caps );

	return 0;
}

// "pamcap_user_t" routines ///////////////////////////////////////////////
void pamcap_user_init( pamcap_user_t * user )
{
	if ( user == NULL )
		return;

	memset( user->username, 0, MAX_USERNAME );
	memset( user->rolename, 0, MAX_ROLENAME );
	user->next = NULL;
}

void pamcap_user_freelist( pamcap_user_t * usertop )
{
	pamcap_user_t * ptr1, * ptr2;

	for ( ptr1 = usertop; ptr1 != NULL; ptr1 = ptr2 ) {
		ptr2 = ptr1->next;
		free( ptr1 );
	}
}

pamcap_user_t * pamcap_user_findlist( pamcap_user_t * usertop,
					   const char * username )
{
	pamcap_user_t * tmp = NULL;

	for ( tmp = usertop; tmp != NULL; tmp = tmp->next )
		if ( strcmp(tmp->username, username) == 0 )
			break;

	return tmp;
}

int pamcap_user_appendlist( pamcap_user_t ** usertop, pamcap_user_t * new )
{
	pamcap_user_t * up;

	if ( usertop == NULL )
		return -1;

	if ( new == NULL )
		return 0;

	// enforce uniqueness
	if ( pamcap_user_findlist(*usertop, new->username) != NULL ) {
		syslog( LOG_ERR, "duplicate user (%s) ignored", new->username );
		return -1;
	}

	if ( *usertop == NULL )
		*usertop = new;
	else {
		for ( up = *usertop; up->next != NULL; up = up->next )
			;;
		up->next = new;
	}

	return 0;
}

int pamcap_user_verify( pamcap_user_t * user )
{
	struct passwd * pp = NULL;

	if ( user == NULL )
		return -1;

	if ( strlen(user->username) == 0            ||
	     strlen(user->username) > MAX_USERNAME  ||
	     strlen(user->rolename) == 0            ||
	     strlen(user->rolename) > MAX_ROLENAME )
		return -1;

	if ( strncmp(user->username, "*", 1) == 0 )
		return 0;

	if ( (pp = getpwnam(user->username)) == NULL )
		return -1;

	return 0;
}

// "pamcap_group_t" routines ///////////////////////////////////////////////
void pamcap_group_init( pamcap_group_t * group )
{
	if ( group == NULL )
		return;

	memset( group->groupname, 0, MAX_GROUPNAME );
	memset( group->rolename, 0, MAX_ROLENAME );
	group->next = NULL;
}

void pamcap_group_freelist( pamcap_group_t * grouptop )
{
	pamcap_group_t * ptr1, * ptr2;

	for ( ptr1 = grouptop; ptr1 != NULL; ptr1 = ptr2 ) {
		ptr2 = ptr1->next;
		free( ptr1 );
	}
}

pamcap_group_t * pamcap_group_findlist( pamcap_group_t * grouptop,
					const char * groupname )
{
	pamcap_group_t * tmp = NULL;

	for ( tmp = grouptop; tmp != NULL; tmp = tmp->next )
		if ( strcmp(tmp->groupname, groupname) == 0 )
			break;

	return tmp;
}

int pamcap_group_appendlist( pamcap_group_t ** grouptop, pamcap_group_t * new )
{
	pamcap_group_t * up;

	if ( grouptop == NULL )
		return -1;

	if ( new == NULL )
		return 0;

	// enforce uniqueness
	if ( pamcap_group_findlist(*grouptop, new->groupname) != NULL ) {
		syslog( LOG_ERR, "duplicate group (%s) ignored",
			new->groupname );
		return -1;
	}

	if ( *grouptop == NULL )
		*grouptop = new;
	else {
		for ( up = *grouptop; up->next != NULL; up = up->next )
			;;
		up->next = new;
	}

	return 0;
}

pamcap_group_t * pamcap_group_findbyuser( pamcap_group_t * grouptop,
					  const char * username )
{
	pamcap_group_t * tmp = NULL;
	struct group * gp = NULL;
	char ** gmem = NULL;

	for ( tmp = grouptop; tmp != NULL; tmp = tmp->next ) {
		if ( (gp = getgrnam(tmp->groupname)) != NULL ) {
			for ( gmem = gp->gr_mem; gmem != NULL &&
			      *gmem != NULL && **gmem != '\0'; gmem++ ) {
				if ( strncmp(username, *gmem, MAX_USERNAME)
				     == 0 )
					break;
			}
			if ( gmem != NULL && *gmem != NULL && **gmem != '\0' )
				break;
		}
	}

	return tmp;
}

int pamcap_group_verify( pamcap_group_t * group )
{
	struct group * gp = NULL;

	if ( group == NULL )
		return -1;

	if ( strlen(group->groupname) == 0             ||
	     strlen(group->groupname) > MAX_GROUPNAME  ||
	     strlen(group->rolename)  == 0             ||
	     strlen(group->rolename)  > MAX_ROLENAME )
		return -1;

	if ( (gp = getgrnam(group->groupname)) == NULL )
		return -1;

	return 0;
}

/*
 * [ pamcap_set_by_user ]
 *
 * steps:
 * 	1 - check if the user exists
 * 	2 - if uid == 0, bail out with success. root has privilidges already.
 * 	3 - parse the conf file
 * 	4 - find rolename by
 * 	    1 - username,
 * 	    2 - group membership,
 * 	    3 - then user '*'
 * 	5 - attempt to set capabilities based on the role
 */
int pamcap_set_by_user( const char * username, pamcap_obj_t * obj )
{
	struct passwd * pwd = NULL;
	pamcap_role_t  * rptr = NULL;
	pamcap_user_t  * uptr = NULL;
	pamcap_group_t * gptr = NULL;
	int rval = -1;
	cap_t caps = NULL;
	char rolename[MAX_ROLENAME];
	char capabilities[MAX_CAPTEXT];

	if ( username == NULL || obj == NULL || obj->conf == NULL )
		goto __out;

	memset( rolename, 0, MAX_ROLENAME );
	memset( capabilities, 0, MAX_CAPTEXT );

	// [1]
	if ( (pwd = getpwnam(username)) != NULL )
		// [2]
		if ( pwd->pw_uid == 0 ) {
			rval = 0;
			goto __out;
		}

	// [3]
	if ( pamcap_conf_parse(obj) != 0 ) {
		syslog( LOG_ERR, "error parsing conf file <%s>", obj->conf );
		goto __out;
	}

	// [4]
	// [4.1]
	uptr = pamcap_user_findlist( obj->tbl_users, username );
	if ( uptr != NULL )
		strncpy( rolename, uptr->rolename, MAX_ROLENAME-1 );

	if ( obj->debug )
		syslog( LOG_INFO, "user declaration for %s %s", username,
			uptr == NULL ? "not found" : "found" );

	// [4.2]
	if ( strlen(rolename) == 0 ) {
		if ( (gptr = pamcap_group_findbyuser(obj->tbl_groups, username))
		     != NULL )
			strncpy( rolename, gptr->rolename, MAX_ROLENAME-1 );

		if ( obj->debug )
			syslog( LOG_INFO,
				"group declaration with member %s %s %s%s%s",
				username,
				gptr == NULL ? "not found" : "found",
				gptr == NULL ? "" : "(",
				gptr == NULL ? "" : gptr->groupname,
				gptr == NULL ? "" : ")" );
	}

	// [4.3]
	if ( strlen(rolename) == 0 ) {
		if ( (uptr = pamcap_user_findlist(obj->tbl_users, "*"))
		     != NULL )
			strncpy( rolename, uptr->rolename, MAX_ROLENAME-1 );

		if ( obj->debug )
			syslog( LOG_INFO, "user declaration for '*' %s",
				uptr == NULL ? "not found" : "found" );
	}

	if ( strlen(rolename) == 0 ) {
		if ( obj->debug )
			syslog( LOG_INFO, "couldn't find config for user %s",
				username );
		goto __out;
	}

	rptr = pamcap_role_findlist( obj->tbl_roles, rolename );
	if ( rptr == NULL ) {
		syslog( LOG_ERR, "couldn't find config for role %s",
			uptr->rolename );
		goto __out;
	}

	if ( strncmp(rolename, rptr->rolename, MAX_ROLENAME) != 0 ) {
		syslog( LOG_ERR, "failed paranoid check" );
		goto __out;
	}

	strncpy( capabilities, rptr->capabilities,
		 MAX_CAPTEXT-strlen(CAP_EXTRA)-1 );
	if ( obj->append != NULL ) {
		APPEND( capabilities, " ", MAX_CAPTEXT-1 );
		APPEND( capabilities, obj->append, MAX_CAPTEXT-1 );
	}
	APPEND( capabilities, " ", MAX_CAPTEXT-1 );
	APPEND( capabilities, CAP_EXTRA, MAX_CAPTEXT-1 );

	if ( (caps = cap_from_text(capabilities)) == NULL ) {
		syslog( LOG_ERR, "couldn't cap_from_text" );
		goto __out;
	}

	// [5]
	if ( cap_set_proc(caps) != 0 ) {
		syslog( LOG_ERR,
			"couldn't set caps (%s) for user %s, role %s : %s",
			capabilities, username, rolename,
			strerror(errno) );
		goto __out;
	} else if ( obj->debug )
		syslog( LOG_INFO, "set caps (%s) for user %s, role %s",
			rptr->capabilities, username, rolename );

	rval = 0;
__out:
	if ( obj->tbl_roles != NULL )
		pamcap_role_freelist( obj->tbl_roles );

	if ( obj->tbl_users != NULL )
		pamcap_user_freelist( obj->tbl_users );

	if ( obj->tbl_groups != NULL )
		pamcap_group_freelist( obj->tbl_groups );

	if ( caps != NULL ) {
		cap_free( caps );
		caps = NULL;
	}

	return rval;
}

void pamcap_set_default( pamcap_obj_t * obj )
{
	cap_t caps = NULL;
	char capabilities[MAX_CAPTEXT];

	if ( obj != NULL && obj->debug )
		syslog( LOG_INFO, "clearing all capabilities" );

	memset( capabilities, 0, MAX_CAPTEXT );

	strncpy( capabilities, "all=", 4 );
	if ( obj->append != NULL ) {
		APPEND( capabilities, " ", MAX_CAPTEXT-1 );
		APPEND( capabilities, obj->append, MAX_CAPTEXT-1 );
	}
	APPEND( capabilities, " ", MAX_CAPTEXT-1 );
	APPEND( capabilities, CAP_EXTRA, MAX_CAPTEXT-1 );

	caps = cap_from_text( capabilities );
	cap_set_proc( caps );
	cap_free( caps );

	return;
}

// config file routines ///////////////////////////////////////////////////
int pamcap_conf_check( pamcap_obj_t * obj )
{
	struct stat buf;

	if ( obj == NULL || obj->conf == NULL )
		return -1;

	memset( &buf, 0, sizeof(buf) );
	if ( stat(obj->conf, &buf) != 0 ) {
		syslog( LOG_ERR, "error stat'ing conf file <%s>: %s",
			obj->conf, strerror(errno) );
		return -1;
	}

	if ( buf.st_uid != 0 )
		return -1;

	if ( !S_ISREG(buf.st_mode) )
		return -1;

	if ( buf.st_mode & CONF_BAD_PERMS )
		return -1;

	return 0;
}

/*
 * the attempted format of the capability.conf file is:
 *
 * role    <rolename>      <cap_list>
 * group   <groupname>     <rolename>
 * user    <username>      <rolename>
 *
 * where <cap_list> conforms to cap_from_text(3) syntax.
 *
 * '#' are comments and \'s can be used to continue lines.
 *
 * also: roles must be declared before they are used in a 'user'
 * or 'group' declaration, otherwise 'role', 'user', and 'group'
 * declarations can be intermingled, though it's probably cleaner
 * not to do so.
 */
int pamcap_conf_parse( pamcap_obj_t * obj )
{
	int rval = -1;
	FILE * conf = NULL;
	int i;
	char line[MAX_LINE];
	pamcap_role_t  * rptr = NULL;
	pamcap_user_t  * uptr = NULL;
	pamcap_group_t * gptr = NULL;
	int ps_cmd; // parser state variable
#define	PS_CMD_ROLE_KEY		1
#define	PS_CMD_ROLE_VAL		2
#define	PS_CMD_USER_KEY		3
#define	PS_CMD_USER_VAL		4
#define	PS_CMD_GROUP_KEY	5
#define	PS_CMD_GROUP_VAL	6

	if ( obj == NULL )
		goto __out;

	if ( obj->tbl_roles != NULL )
		pamcap_role_freelist( obj->tbl_roles );

	if ( obj->tbl_users != NULL )
		pamcap_user_freelist( obj->tbl_users );

	if ( obj->tbl_groups != NULL )
		pamcap_group_freelist( obj->tbl_groups );

	if ( pamcap_conf_check(obj) != 0 ) {
		syslog( LOG_ERR, "error validating conf file <%s>",
			obj->conf );
		goto __out;
	}

	if ( (conf = fopen(obj->conf, "r")) == NULL ) {
		syslog( LOG_ERR, "error fopen'ing conf file <%s>: %s",
			obj->conf, strerror(errno) );
		goto __out;
	}

	// OK, up until now I've tried to follow the CodingStyle
	// document for the Linux kernel. But I will break here only
	// because this is a big parsing loop. So here it's one tab,
	// and then 4 spaces for each subsequent level.
	i = 1;
	ps_cmd = 0;
	while( fgets(line, MAX_LINE, conf) ) {
	    char * cp;
	    int len = 0;

	    line[MAX_LINE - 1] = '\0';

	    if ( (len = strlen(line)) == 0 )
	        goto __continue;

	    if ( line[0] == '\0' || line[0] == '#' )
	        goto __continue;

	    cp = strtok( line, WHITESPACE );
	    if ( cp == NULL || *cp == '\0' )
	        goto __continue;

/*
{
	pamcap_role_t  * rp;
	pamcap_user_t  * up;
	pamcap_group_t * gp;

	for ( rp = obj->tbl_roles; rp != NULL; rp = rp->next )
		syslog( LOG_INFO, "role %s:%s", rp->rolename,
			rp->capabilities );

	for ( up = obj->tbl_users; up != NULL; up = up->next )
		syslog( LOG_INFO, "user %s:%s", up->username,
			up->rolename );

	for ( gp = obj->tbl_groups; gp != NULL; gp = gp->next )
		syslog( LOG_INFO, "group %s:%s", gp->groupname,
			gp->rolename );
} while(0);
*/

	    if ( ! ps_cmd && strcmp(cp, "role") == 0 ) {
	        ps_cmd = PS_CMD_ROLE_KEY;
	        if ( (rptr = malloc(sizeof(pamcap_role_t)))
	             == NULL ) {
	            syslog( LOG_ERR, "malloc role problem" );
	            goto __out;
	        }
	        pamcap_role_init( rptr );
	        cp = strtok( NULL, WHITESPACE );
	        if ( cp == NULL || *cp == '\0' )
	            goto __out;
	    } else if ( ! ps_cmd && strcmp(cp, "user") == 0 ) {
	        ps_cmd = PS_CMD_USER_KEY;
	        if ( (uptr = malloc(sizeof(pamcap_user_t)))
	             == NULL ) {
	            syslog( LOG_ERR, "malloc user problem" );
	            goto __out;
	        }
	        pamcap_user_init( uptr );
	        cp = strtok( NULL, WHITESPACE );
	        if ( cp == NULL || *cp == '\0' )
	            goto __out;
	    } else if ( ! ps_cmd && strcmp(cp, "group") == 0 ) {
	        ps_cmd = PS_CMD_GROUP_KEY;
	        if ( (gptr = malloc(sizeof(pamcap_group_t)))
	             == NULL ) {
	            syslog( LOG_ERR, "malloc group problem" );
	            goto __out;
	        }
	        pamcap_group_init( gptr );
	        cp = strtok( NULL, WHITESPACE );
	        if ( cp == NULL || *cp == '\0' )
	            goto __out;
	    }

	    if ( ps_cmd == PS_CMD_ROLE_KEY ) {
	        if ( cp == NULL || (! isalnum(*cp) && *cp != '\\') ) {
				syslog( LOG_ERR, "bad line: %d", i );
				goto __out;
	        } else if ( isalnum(*cp) ) {
	            strncpy( rptr->rolename, cp, MAX_ROLENAME - 1 );
	            ps_cmd = PS_CMD_ROLE_VAL;
	            cp = strtok( NULL, WHITESPACE );
	            if ( cp == NULL || *cp == '\0' )
	                goto __out;
	        } else if ( *cp == '\\' )
	            goto __continue;
	    }

	    if ( ps_cmd == PS_CMD_ROLE_VAL ) {
__role_val:
	        if ( cp == NULL || (! isalnum(*cp) && strcmp(cp, "=") != 0 &&
	             strcmp(cp, "\\") != 0) ) {
	            syslog( LOG_ERR, "bad line: %d", i );
	            goto __out;
	        } else if ( isalnum(*cp) || strcmp(cp, "=") == 0 ) {
	            if ( index(cp, '=') == NULL && index(cp, '+') == NULL &&
	                 index(cp, '-') == NULL && index(cp, ',') == NULL ) {
	                // it's probably a role name
	                pamcap_role_t * tmp_role;
			tmp_role = pamcap_role_findlist( obj->tbl_roles, cp );
			if ( tmp_role == NULL ) {
	                    syslog( LOG_ERR, "unknown role: %s", cp );
	                    goto __out;
	                }
	                APPEND( rptr->capabilities, tmp_role->capabilities,
	                        MAX_CAPTEXT-1 );
	                APPEND( rptr->capabilities, " ", MAX_CAPTEXT-1 );
	            } else {
	                // it's probably a capability list
	                APPEND( rptr->capabilities, cp, MAX_CAPTEXT-1 );
	                APPEND( rptr->capabilities, " ", MAX_CAPTEXT-1 );
	            }
	            cp = strtok( NULL, WHITESPACE );
	            if ( cp == NULL || *cp == '\0' ) {
	                ps_cmd = 0;
			if ( pamcap_role_verify(rptr) == 0 )
	                    pamcap_role_appendlist( &(obj->tbl_roles), rptr );
			else {
	                    syslog( LOG_ERR, "couldn't verify role %s: ignored",
	                            rptr->rolename );
	                    free( rptr );
	                }
	                rptr = NULL;
	            } else
	                goto __role_val;
	        }
	        goto __continue;
	    }

	    if ( ps_cmd == PS_CMD_USER_KEY ) {
	        if ( cp == NULL || (! isalnum(*cp) && strcmp(cp, "\\") != 0 &&
	             strcmp(cp, "*") != 0) ) {
	            syslog( LOG_ERR, "bad line: %d", i );
	            goto __out;
	        } else if ( isalnum(*cp) || strcmp(cp, "*") == 0 ) {
	            strncpy( uptr->username, cp, MAX_USERNAME - 1 );
	            ps_cmd = PS_CMD_USER_VAL;
	            cp = strtok( NULL, WHITESPACE );
	            if ( cp == NULL || *cp == '\0' )
	                goto __out;
	        } else if ( strcmp(cp, "\\") == 0 )
	            goto __continue;
	    }

	    if ( ps_cmd == PS_CMD_USER_VAL ) {
	        if ( cp == NULL || (! isalnum(*cp) &&
	             strcmp(cp, "\\") != 0) ) {
	            syslog( LOG_ERR, "bad line: %d", i );
	            goto __out;
	        } else if ( isalnum(*cp) ) {
	            if ( pamcap_role_findlist(obj->tbl_roles, cp) == NULL ) {
	                syslog( LOG_ERR, "undefined role: %s", cp );
	                goto __out;
	            }
	            strncpy( uptr->rolename, cp, MAX_ROLENAME - 1 );
	            ps_cmd = 0;
		    if ( pamcap_user_verify(uptr) == 0 )
	                pamcap_user_appendlist( &(obj->tbl_users), uptr );
		    else {
	                syslog( LOG_ERR, "couldn't verify user %s: ignored",
	                        uptr->username );
	                free( uptr );
		    }
	            uptr = NULL;
	        }
	        goto __continue;
	    }

	    if ( ps_cmd == PS_CMD_GROUP_KEY ) {
	        if ( cp == NULL || (! isalnum(*cp) && strcmp(cp, "\\") != 0 &&
	             strcmp(cp, "*") != 0) ) {
	            syslog( LOG_ERR, "bad line: %d", i );
	            goto __out;
	        } else if ( isalnum(*cp) || strcmp(cp, "*") == 0 ) {
	            strncpy( gptr->groupname, cp, MAX_GROUPNAME - 1 );
	            ps_cmd = PS_CMD_GROUP_VAL;
	            cp = strtok( NULL, WHITESPACE );
	            if ( cp == NULL || *cp == '\0' )
	                goto __out;
	        } else if ( strcmp(cp, "\\") == 0 )
	            goto __continue;
	    }

	    if ( ps_cmd == PS_CMD_GROUP_VAL ) {
	        if ( cp == NULL || (! isalnum(*cp) &&
	             strcmp(cp, "\\") != 0) ) {
	            syslog( LOG_ERR, "bad line: %d", i );
	            goto __out;
	        } else if ( isalnum(*cp) ) {
	            if ( pamcap_role_findlist(obj->tbl_roles, cp) == NULL ) {
	                syslog( LOG_ERR, "undefined role: %s", cp );
	                goto __out;
	            }
	            strncpy( gptr->rolename, cp, MAX_ROLENAME - 1 );
	            ps_cmd = 0;
	            if ( pamcap_group_verify(gptr) == 0 )
	                pamcap_group_appendlist( &(obj->tbl_groups), gptr );
		    else {
	                syslog( LOG_ERR, "couldn't verify group %s: ignored",
	                        gptr->groupname );
	                free( gptr );
		    }
	            gptr = NULL;
	        }
	        goto __continue;
	    }

__continue:
		i++;
		memset( line, 0, MAX_LINE );
	}
	// End of CodingStyle alteration.

	rval = 0;
__out:
	if ( rptr != NULL )
	    free( rptr );
	if ( uptr != NULL )
	    free( uptr );
	if ( gptr != NULL )
	    free( gptr );

	if ( conf != NULL ) {
		fclose( conf );
		conf = NULL;
	}

	if ( rval != 0 ) {
		pamcap_role_freelist( obj->tbl_roles );
		obj->tbl_roles = NULL;
		pamcap_user_freelist( obj->tbl_users );
		obj->tbl_users = NULL;
		pamcap_group_freelist( obj->tbl_groups );
		obj->tbl_groups = NULL;
	}

	return rval;
}

